/*
								+---------------------------------+
								|                                 |
								| *** Discrete cos. transform *** |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|              DCT.h              |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 08-22-2006
 *
 *	fixed error in CDCT8_1D and CFastDCT8_2D::Forward
 *	everything is working well now
 *
 */

#ifndef __DCT_INCLUDED
#define __DCT_INCLUDED

extern const float f_pi;// = 3.14159265358979323846264338327950288419716939937510f

/*
 *	template <class T, const int n_offset, const T m1, const T m2, const T m3, const T m4>
 *		const int n_mul_shift>
 *	class CFDCT8_1D
 *		- original code by Yann Guidon in article "Optimised Winograd DCT for the F-CPU Core 0"
 *		- T is data type operations are performed on (int16 should be sufficient)
 *		- n_mul_shift is bit-shift after multiplication (use 0 for T = float, 8 for int16, etc)
 *		- n_offset is offset between source samples (1 for line DCT, 8 for row DCT)
 *		- m1 to m4 are precalculated constants in required data type:
 *	 		m1 = cos(4pi/16)
 *			m2 = cos(6pi/16)
 *			m3 = cos(2pi/16) - cos(6pi/16)
 *			m4 = cos(2pi/16) + cos(6pi/16)
 *			(0.707107f, 0.382683f, 0.541196f, 1.306563f)
 *		- can operate in-sit (p_dest can be equal to p_src)
 */
template <class T, const int n_offset, const T m1, const T m2, const T m3, const T m4,
	const int n_mul_shift>
class CFDCT8_1D {
public:
	inline void operator ()(T *p_dest, const T *p_src)
	{
		T n_tmp_7 = p_src[0 * n_offset] + p_src[7 * n_offset];
		T n_tmp_6 = p_src[1 * n_offset] + p_src[6 * n_offset];
		T n_tmp_4 = p_src[3 * n_offset] - p_src[4 * n_offset];
		T n_tmp_5 = p_src[1 * n_offset] - p_src[6 * n_offset];
		T n_tmp_9 = p_src[2 * n_offset] + p_src[5 * n_offset];
		T n_tmp_0 = p_src[3 * n_offset] + p_src[4 * n_offset];
		T n_tmp_3 = p_src[2 * n_offset] - p_src[5 * n_offset];
		T n_tmp_8 = p_src[0 * n_offset] - p_src[7 * n_offset];

		T n_tmp_1 = n_tmp_7 + n_tmp_0;
		T n_tmp_2 = n_tmp_6 - n_tmp_9;
		n_tmp_4 = n_tmp_4 + n_tmp_3;
		n_tmp_3 = n_tmp_5 + n_tmp_3;
		n_tmp_7 = n_tmp_7 - n_tmp_0;
		n_tmp_5 = n_tmp_5 + n_tmp_8;
		n_tmp_9 = n_tmp_6 + n_tmp_9;

		n_tmp_3 = (T)((m1 * n_tmp_3) / (1 << n_mul_shift));
		n_tmp_2 = n_tmp_2 + n_tmp_7;
		n_tmp_6 = n_tmp_4 - n_tmp_5;

		n_tmp_4 = (T)((m3 * n_tmp_4) / (1 << n_mul_shift));
		n_tmp_5 = (T)((m4 * n_tmp_5) / (1 << n_mul_shift));
		n_tmp_2 = (T)((m1 * n_tmp_2) / (1 << n_mul_shift));
		n_tmp_6 = (T)((m2 * n_tmp_6) / (1 << n_mul_shift));

		p_dest[0 * n_offset] = n_tmp_1 + n_tmp_9;
		p_dest[4 * n_offset] = n_tmp_1 - n_tmp_9;

		n_tmp_1 = n_tmp_3 + n_tmp_8;
		n_tmp_3 = n_tmp_8 - n_tmp_3;
		p_dest[2 * n_offset] = n_tmp_7 + n_tmp_2;
		n_tmp_4 = n_tmp_4 + n_tmp_6;
		p_dest[6 * n_offset] = n_tmp_7 - n_tmp_2;
		n_tmp_2 = n_tmp_5 + n_tmp_6;

		p_dest[3 * n_offset] = n_tmp_3 - n_tmp_4;
		p_dest[5 * n_offset] = n_tmp_3 + n_tmp_4;
		p_dest[1 * n_offset] = n_tmp_1 + n_tmp_2;
		p_dest[7 * n_offset] = n_tmp_1 - n_tmp_2;
	}
};

/*
 *	template <class T, const int n_offset, const T m1, const T m2, const T m3, const T m4,
 *		const int n_mul_shift>
 *	class CFDCT8_1D
 *		- somewhat (~50% faster) optimised version of algorithm used in Thomas G. Lane's jpeglib
 *		- T is data type operations are performed on (int16 should be sufficient)
 *		- n_mul_shift is bit-shift after multiplication (use 0 for T = float, 8 for int16, etc)
 *		- n_offset is offset between source samples (1 for line DCT, 8 for row DCT)
 *		- m1 to m4 are precalculated constants in required data type:
 *	 		m1 = 1.414213562f
 *			m2 = 1.847759065f
 *			m3 = 1.082392200f
 *			m4 = -2.613125930f
 *		- can operate in-sit (p_dest can be equal to p_src)
 */
template <class T, const int n_offset, const T m1, const T m2, const T m3, const T m4,
	const int n_mul_shift>
class CIDCT8_1D {
public:
	inline void operator ()(T *p_dest, const T *p_src)
	{
		T n_tmp_4 = p_src[0 * n_offset] + p_src[4 * n_offset];
		T n_tmp_5 = p_src[0 * n_offset] - p_src[4 * n_offset];
		T n_tmp_7 = p_src[2 * n_offset] + p_src[6 * n_offset];
		T n_tmp_6 = (T)(((p_src[2 * n_offset] - p_src[6 * n_offset]) * m1) /
			(1 << n_mul_shift)) - n_tmp_7;
		T n_tmp_9 = p_src[5 * n_offset] + p_src[3 * n_offset];
		T n_tmp_8 = p_src[1 * n_offset] + p_src[7 * n_offset];

		T n_tmp_0 = n_tmp_4 + n_tmp_7;
		T n_tmp_3 = n_tmp_4 - n_tmp_7;
		T n_tmp_1 = n_tmp_5 + n_tmp_6;
		T n_tmp_2 = n_tmp_5 - n_tmp_6;

		n_tmp_5 = (T)(((n_tmp_8 - n_tmp_9) * m1) / (1 << n_mul_shift));
		n_tmp_9 += n_tmp_8;

		p_dest[0 * n_offset] = (n_tmp_0 + n_tmp_9);
		p_dest[7 * n_offset] = (n_tmp_0 - n_tmp_9);

		n_tmp_0 = p_src[5 * n_offset] - p_src[3 * n_offset];
		n_tmp_8 = p_src[1 * n_offset] - p_src[7 * n_offset];

		n_tmp_7 = (T)(((n_tmp_0 + n_tmp_8) * m2) / (1 << n_mul_shift));
		n_tmp_4 = (T)((n_tmp_8 * m3) / (1 << n_mul_shift)) - n_tmp_7;
		n_tmp_6 = (T)((n_tmp_0 * m4) / (1 << n_mul_shift)) + n_tmp_7;

		n_tmp_6 -= n_tmp_9;
		n_tmp_5 -= n_tmp_6;
		n_tmp_4 += n_tmp_5;

		p_dest[1 * n_offset] = (n_tmp_1 + n_tmp_6);
		p_dest[6 * n_offset] = (n_tmp_1 - n_tmp_6);
		p_dest[2 * n_offset] = (n_tmp_2 + n_tmp_5);
		p_dest[5 * n_offset] = (n_tmp_2 - n_tmp_5);
		p_dest[4 * n_offset] = (n_tmp_3 + n_tmp_4);
		p_dest[3 * n_offset] = (n_tmp_3 - n_tmp_4);
	}
};

class CReferenceDCT8_2D {
public:
	static void Forward(float *p_dest, const float *p_src);
	static void Inverse(float *p_dest, const float *p_src);
};

class CFastDCT8_2D {
protected:
	static const float m_p_prescale_table[8];

public:
	static const float *p_PrescaleTable();
	// AA&N family DCT need pre-scaled coefficients (returns pointer to array of 8 floats,
	// 8x8 block is multiplied by p_PrescaleTable()[i / 8] * p_PrescaleTable()[i % 8]
	// where i = <0, 63>)

#ifdef WIN32 // g++ forbids float to be template argument
	static void Inverse(float *p_dest, const float *p_src);
	static void Forward(float *p_dest, const float *p_src);
#endif

	static void Inverse(int *p_dest, const int *p_src);
	static void Inverse(short *p_dest, const short *p_src);
	static void Forward(int *p_dest, const int *p_src);
	static void Forward(short *p_dest, const short *p_src); // faster, integer arithmetic versions
};

#endif //__DCT_INCLUDED
