/*
								+---------------------------------+
								|                                 |
								|  ***   Tiny jpeg decoder   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            Jpeg.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <crtdbg.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "Jpeg.h"
#include "DCT.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

/*
 *								=== CJpegDecoder ===
 */

char CJpegDecoder::p_zig_indices[/*64*/63] = {
   /*0,*/1,  8, 16,  9,  2,  3, 10,
	17, 24, 32, 25, 18, 11,  4,  5,
	12, 19, 26, 33, 40, 48, 41, 34,
	27, 20, 13,  6,  7, 14, 21, 28,
	35, 42, 49, 56, 57, 50, 43, 36,
	29, 22, 15, 23, 30, 37, 44, 51,
	58, 59, 52, 45, 38, 31, 39, 46,
	53, 60, 61, 54, 47, 55, 62, 63
};
// reverse zig-zag indices

enum {
	marker_SOF0_HuffBaseline_DCT = 0xffc0,
	marker_SOF1_HuffExtendedSequential_DCT = 0xffc1,
	marker_SOF2_HuffProgressive_DCT = 0xffc2,
	marker_SOF3_HuffLossless = 0xffc3,
	// start of frame markers, non-differential, huffman coding

	marker_SOF5_HuffDiffSequential_DCT = 0xffc5,
	marker_SOF6_HuffDiffProgressive_DCT = 0xffc6,
	marker_SOF7_HuffDiffLossless = 0xffc7,
	// start of frame markers, differential, huffman coding

	marker_SOF8_arith_reserved = 0xffc8,
	marker_SOF9_arith_extended_sequential_DCT = 0xffc9,
	marker_SOF10_arith_progressive_DCT = 0xffca,
	marker_SOF11_arith_lossless = 0xffcb,
	// start of frame markers, non-differential, arithmetic coding

	marker_SOF13_arith_diff_sequential_DCT = 0xffcd,
	marker_SOF14_arith_diff_progressive_DCT = 0xffce,
	marker_SOF15_arith_diff_lossless = 0xffcf,
	// start of frame markers, differential, arithmetic coding

	marker_HuffmanTable = 0xffc4,
	// huffman table specification

	marker_ArithCodingCondBlock = 0xffcc,
	// arithmetic coding conditioning specification

	marker_RestartMod0 = 0xffd0,
	marker_RestartMod1 = 0xffd1,
	marker_RestartMod2 = 0xffd2,
	marker_RestartMod3 = 0xffd3,
	marker_RestartMod4 = 0xffd4,
	marker_RestartMod5 = 0xffd5,
	marker_RestartMod6 = 0xffd6,
	marker_RestartMod7 = 0xffd7,
	// restart interval termination

	marker_StartImage = 0xffd8,
	marker_EndImage = 0xffd9,
	marker_StartScan = 0xffda,
	marker_DefineQuantTables = 0xffdb,
	marker_DefineNumberLines = 0xffdc,
	marker_DefineRestartInterval = 0xffdd,
	marker_DefineHierarchicalProgression = 0xffde,
	marker_ExpandRefComps = 0xffdf,
	// other markers

	marker_ReservedAppSeg_0 = 0xffe0, // JFIF (jpeg file interchange format - the used format)
	marker_ReservedAppSeg_1 = 0xffe1,
	marker_ReservedAppSeg_2 = 0xffe2,
	marker_ReservedAppSeg_3 = 0xffe3,
	marker_ReservedAppSeg_4 = 0xffe4,
	marker_ReservedAppSeg_5 = 0xffe5,
	marker_ReservedAppSeg_6 = 0xffe6,
	marker_ReservedAppSeg_7 = 0xffe7,
	marker_ReservedAppSeg_8 = 0xffe8,
	marker_ReservedAppSeg_9 = 0xffe9,
	marker_ReservedAppSeg_a = 0xffea,
	marker_ReservedAppSeg_b = 0xffeb,
	marker_ReservedAppSeg_c = 0xffec,
	marker_ReservedAppSeg_d = 0xffed,
	marker_ReservedAppSeg_e = 0xffee,
	marker_ReservedAppSeg_f = 0xffef,

	marker_JPEG_ReservedExt_0 = 0xfff0,
	marker_JPEG_ReservedExt_1 = 0xfff1,
	marker_JPEG_ReservedExt_2 = 0xfff2,
	marker_JPEG_ReservedExt_3 = 0xfff3,
	marker_JPEG_ReservedExt_4 = 0xfff4,
	marker_JPEG_ReservedExt_5 = 0xfff5,
	marker_JPEG_ReservedExt_6 = 0xfff6,
	marker_JPEG_ReservedExt_7 = 0xfff7,
	marker_JPEG_ReservedExt_8 = 0xfff8,
	marker_JPEG_ReservedExt_9 = 0xfff9,
	marker_JPEG_ReservedExt_a = 0xfffa,
	marker_JPEG_ReservedExt_b = 0xfffb,
	marker_JPEG_ReservedExt_c = 0xfffc,
	marker_JPEG_ReservedExt_d = 0xfffd,

	marker_Comment = 0xfffe,

	marker_Temp = 0xff01,

	marker_Reserved_min = 0xff02,
	marker_Reserved_max = 0xffbf
};
// marker codes

const CJpegDecoder::TBlockAction CJpegDecoder::m_p_action_table[] = {
	{marker_SOF0_HuffBaseline_DCT, Read_FrameHeader}, // -------------
	{marker_SOF1_HuffExtendedSequential_DCT, Read_Unsupported},
	{marker_SOF2_HuffProgressive_DCT, Read_Unsupported},
	{marker_SOF3_HuffLossless, Read_Unsupported},
	{marker_SOF5_HuffDiffSequential_DCT, Read_Unsupported},
	{marker_SOF6_HuffDiffProgressive_DCT, Read_Unsupported},
	{marker_SOF7_HuffDiffLossless, Read_Unsupported},
	{marker_SOF8_arith_reserved, Read_Unsupported},
	{marker_SOF9_arith_extended_sequential_DCT, Read_Unsupported},
	{marker_SOF10_arith_progressive_DCT, Read_Unsupported},
	{marker_SOF11_arith_lossless, Read_Unsupported},
	{marker_SOF13_arith_diff_sequential_DCT, Read_Unsupported},
	{marker_SOF14_arith_diff_progressive_DCT, Read_Unsupported},
	{marker_SOF15_arith_diff_lossless, Read_Unsupported},
	{marker_HuffmanTable, Read_HuffmanTables},
	{marker_ArithCodingCondBlock, Read_Unsupported},
	/*{marker_RestartMod0, Read_Skip},
	{marker_RestartMod1, Read_Skip},
	{marker_RestartMod2, Read_Skip},
	{marker_RestartMod3, Read_Skip},
	{marker_RestartMod4, Read_Skip},
	{marker_RestartMod5, Read_Skip},
	{marker_RestartMod6, Read_Skip},
	{marker_RestartMod7, Read_Skip},*/ // 
	{marker_StartImage, Read_SkipMarker},
	{marker_EndImage, Read_SkipMarker},
	{marker_StartScan, Read_StartScan},
	{marker_DefineQuantTables, Read_QuantTables},
	{marker_DefineNumberLines, Read_Unsupported},
	{marker_DefineRestartInterval, Read_RestartInterval},
	{marker_DefineHierarchicalProgression, Read_Unsupported},
	{marker_ExpandRefComps, Read_Skip},
	{marker_ReservedAppSeg_0, Read_JFIF_Header},
	{marker_ReservedAppSeg_1, Read_Skip},
	{marker_ReservedAppSeg_2, Read_Skip},
	{marker_ReservedAppSeg_3, Read_Skip},
	{marker_ReservedAppSeg_4, Read_Skip},
	{marker_ReservedAppSeg_5, Read_Skip},
	{marker_ReservedAppSeg_6, Read_Skip},
	{marker_ReservedAppSeg_7, Read_Skip},
	{marker_ReservedAppSeg_8, Read_Skip},
	{marker_ReservedAppSeg_9, Read_Skip},
	{marker_ReservedAppSeg_a, Read_Skip},
	{marker_ReservedAppSeg_b, Read_Skip},
	{marker_ReservedAppSeg_c, Read_Skip},
	{marker_ReservedAppSeg_d, Read_Skip},
	{marker_ReservedAppSeg_e, Read_Skip},
	{marker_ReservedAppSeg_f, Read_Skip},
	{marker_JPEG_ReservedExt_0, Read_Skip},
	{marker_JPEG_ReservedExt_1, Read_Skip},
	{marker_JPEG_ReservedExt_2, Read_Skip},
	{marker_JPEG_ReservedExt_3, Read_Skip},
	{marker_JPEG_ReservedExt_4, Read_Skip},
	{marker_JPEG_ReservedExt_5, Read_Skip},
	{marker_JPEG_ReservedExt_6, Read_Skip},
	{marker_JPEG_ReservedExt_7, Read_Skip},
	{marker_JPEG_ReservedExt_8, Read_Skip},
	{marker_JPEG_ReservedExt_9, Read_Skip},
	{marker_JPEG_ReservedExt_a, Read_Skip},
	{marker_JPEG_ReservedExt_b, Read_Skip},
	{marker_JPEG_ReservedExt_c, Read_Skip},
	{marker_JPEG_ReservedExt_d, Read_Skip}, // will be treated as unknown blocks
	{marker_Comment, Read_Skip},
	{marker_Temp, Read_Skip},
	{marker_Reserved_min, Read_Skip},
	{marker_Reserved_max, Read_Skip},
};

TBmp *CJpegDecoder::p_Decode_JPEG(FILE *p_fr)
{
	CFileBuffer file_buffer(p_fr);

	m_t_image_data.b_read_image_header = false;
	for(int i = 0; i < 4; i ++) {
		m_t_image_data.p_framebuffer[i] = 0;
		m_t_image_data.n_dc_value[i] = 0;
	}
	m_t_restart.b_enabled = false;
	// clear data that needs to be cleared

	bool b_have_start_image = false;
	while(1) {
		unsigned short n_marker_code;
		if(!file_buffer.Read_Marker(n_marker_code)) {
			for(int i = 0; i < 4; i ++) {
				if(m_t_image_data.p_framebuffer[i])
					delete[] m_t_image_data.p_framebuffer[i];
			}
			return 0;
		}

		if(n_marker_code == marker_StartImage) {
			b_have_start_image = true;

#ifdef __VERBOSE
			printf("encountered start-of-image marker\n");
			// debug
#endif
		} else if(!b_have_start_image)
			continue;
		// skip until reaches marker for start of image

		if(n_marker_code == marker_EndImage)
			break;
		// when encounters marker for end of image, we're done, time to create RGB output

		bool b_found = false;
		for(const TBlockAction *p_action = m_p_action_table, *p_end = m_p_action_table +
		   (sizeof(m_p_action_table) / sizeof(m_p_action_table[0])); p_action < p_end; p_action ++) {
			if(p_action->n_id == n_marker_code) {
				if(!p_action->p_function(this, &file_buffer)) {
					for(int i = 0; i < 4; i ++) {
						if(m_t_image_data.p_framebuffer[i])
							delete[] m_t_image_data.p_framebuffer[i];
					}
					return 0;
				}
				b_found = true;
				break;
			}
		}
#ifndef __STRIP_ERROR_CHECKS
		if(!b_found) {
			for(int i = 0; i < 4; i ++) {
				if(m_t_image_data.p_framebuffer[i])
					delete[] m_t_image_data.p_framebuffer[i];
			}
			return 0;
		}
		// try to process current block
#endif
	}
	// decode image, result should be in m_t_image_data::p_framebuffer

	TBmp *p_bitmap;
	if(!(p_bitmap = new TBmp)) {
		for(int i = 0; i < 4; i ++) {
			if(m_t_image_data.p_framebuffer[i])
				delete[] m_t_image_data.p_framebuffer[i];
		}
		return 0;
	}
	p_bitmap->n_former_bpp = m_t_image_header.n_sample_precision;
	p_bitmap->b_grayscale = m_t_image_header.n_component_num == 1;
	p_bitmap->b_alpha = false; // jpeg files don't have alpha channel
	p_bitmap->n_width = m_t_image_header.n_width;
	p_bitmap->n_height = m_t_image_header.n_height;
	if(!(p_bitmap->p_buffer = new unsigned __int32[p_bitmap->n_width * p_bitmap->n_height])) {
		for(int i = 0; i < 4; i ++) {
			if(m_t_image_data.p_framebuffer[i])
				delete[] m_t_image_data.p_framebuffer[i];
		}
		delete p_bitmap;
		return 0;
	}
	// alloc bitmap structure

	if(m_t_image_header.n_component_num == 1) {
		TComponentInfo *p_Y_component = m_t_image_header.p_component_info;
		// there's just one, we don't have to search

#ifdef __VERBOSE
		printf("rendering grayscale image\n");
#endif

		unsigned __int32 *p_dest_scanline = p_bitmap->p_buffer;
		unsigned char *p_src_buffer = m_t_image_data.p_framebuffer[0];
		for(int y = 0; y < m_t_image_header.n_height / p_Y_component->n_scale_vert; y ++) {
			unsigned __int32 *p_dest_pixel = p_dest_scanline;
			for(int x = 0; x < m_t_image_header.n_width / p_Y_component->n_scale_horiz; x ++) {
				unsigned char n_gray_ub = *p_src_buffer ++ + 128;
				unsigned __int32 n_gray = ((unsigned __int32)n_gray_ub << 16) |
					((unsigned short)n_gray_ub << 8) | (n_gray_ub) | 0xff000000;
				// calc color

				unsigned __int32 *p_end = p_dest_pixel + p_Y_component->n_scale_horiz;
				do {
					*p_dest_pixel ++ = n_gray;
				} while(p_dest_pixel < p_end);
				// fill scanline data, subsample horizontally
			}

			for(int i = 1; i < p_Y_component->n_scale_vert; i ++) {
				memcpy(p_dest_scanline + i * p_bitmap->n_width,
					p_dest_scanline, p_bitmap->n_width * sizeof(unsigned __int32));
			}
			// vertical subscale

			p_dest_scanline += p_Y_component->n_scale_vert * p_bitmap->n_width;
			// move to next destination scanline
		}
		// semi-fast fill algorithm (could write several ones for each subscale,
		// but aim of this decoder is to minimize code size)
	} else /*if(m_t_image_header.n_component_num == 3)*/ {
		_ASSERTE(m_t_image_header.n_component_num == 3);

		memset(p_bitmap->p_buffer, 0, p_bitmap->n_width * p_bitmap->n_height *
			sizeof(unsigned __int32));
		// fill with zeros first, the color components will be or-ed

		short p_YCbCr_to_RGB[4][256]; // 2k table with premultiplied coeffs
		for(short i = 0; i < 256; i ++) {
			p_YCbCr_to_RGB[0][i] = (short)(1.402f *   (i - 128/**/));
			p_YCbCr_to_RGB[1][i] = (short)(1.772f *   (i - 128/**/));
			p_YCbCr_to_RGB[2][i] = (short)(-.34414f * (i - 128/**/));
			p_YCbCr_to_RGB[3][i] = (short)(-.71414f * (i - 128/**/));
		}
		// precalc table to speedup YCbCr to RGB conversion

#ifdef __VERBOSE
		printf("rendering RGB image\n");
#endif

		for(int n = 0; n < 3; n ++) {
			const TComponentInfo *p_component = 0;
			for(TComponentInfo *p_comp = m_t_image_header.p_component_info,
			   *p_end = m_t_image_header.p_component_info + m_t_image_header.n_component_num;
			   p_comp < p_end; p_comp ++) {
				if(p_comp->n_component_id == n + 1) {
					p_component = p_comp;
					break;
				}
			}
#ifndef __STRIP_ERROR_CHECKS
			if(!p_component) {
				for(int i = 0; i < 4; i ++) {
					if(m_t_image_data.p_framebuffer[i])
						delete[] m_t_image_data.p_framebuffer[i];
				}
				delete[] p_bitmap->p_buffer;
				delete p_bitmap;
				return 0;
			}
#endif

#ifdef __VERBOSE
			const char *p_component_name_list[] = {"Y", "Cb", "Cr"};
			printf("\tcomponent %s scaling %d x %d\n", p_component_name_list[n],
				p_component->n_scale_horiz, p_component->n_scale_vert);
			// debug
#endif

			unsigned __int32 *p_dest_scanline = p_bitmap->p_buffer;
			const unsigned char *p_src_buffer = m_t_image_data.p_framebuffer[n];
			int n_framebuffer_width = m_t_image_data.n_framebuffer_width[n];
			for(int y = 0; y < m_t_image_header.n_height; y += p_component->n_scale_vert) {
				for(int i = 0; i < p_component->n_scale_vert; i ++) {
					const unsigned char *p_src_buffer2 = p_src_buffer;
					for(unsigned __int32 *p_dest_pixel = p_dest_scanline;
					   p_dest_pixel <= p_dest_scanline + m_t_image_header.n_width;) {
						unsigned __int32 n_value = ((unsigned __int32)(*p_src_buffer2 ++)) << (n << 3);
						// calc color

						unsigned __int32 *p_end = p_dest_pixel + p_component->n_scale_horiz;
						if(p_end > p_dest_scanline + m_t_image_header.n_width)
							p_end = p_dest_scanline + m_t_image_header.n_width;
						do {
							*p_dest_pixel ++ |= n_value;
						} while(p_dest_pixel < p_end);
						// fill scanline data, subsample horizontally
					}

					if(n == 2) {
						unsigned __int32 *p_dest_pixel = p_dest_scanline;
						unsigned __int32 *p_end = p_dest_scanline + p_bitmap->n_width;
						do {
							unsigned char Y = (unsigned char)(*p_dest_pixel /*& 0xff*/) + 128;
							unsigned char Cb = (unsigned char)((*p_dest_pixel >> 8) /*& 0xff*/) + 128;
							unsigned char Cr = (unsigned char)(*p_dest_pixel >> 16) + 128;

							__int32 R = Y + p_YCbCr_to_RGB[0][Cr];
							if(R < 0)
								R = 0;
							else if(R >= 0xff)
								R = 0xff0000;
							else
								R <<= 16;
							__int32 G = Y + p_YCbCr_to_RGB[2][Cb] + p_YCbCr_to_RGB[3][Cr];
							if(G < 0)
								G = 0;
							else if(G >= 0xff)
								G = 0xff00;
							else
								G <<= 8;
							__int32 B = Y + p_YCbCr_to_RGB[1][Cb];
							if(B < 0)
								B = 0;
							else if(B > 0xff)
								B = 0xff;

							*p_dest_pixel ++ = 0xff000000 | R | G | B;
						} while(p_dest_pixel < p_end);
						// convert from YCbCr to RGB, add alpha channel
					}

					p_dest_scanline += p_bitmap->n_width;
					// move to next destination scanline
				}
				p_src_buffer += n_framebuffer_width;
			}
			// semi-fast fill algorithm (could write several ones for each subscale,
			// but aim of this decoder is to minimize code size)
		}
	}
	// convert from Y or YCbCr to RGB

	for(int i = 0; i < 4; i ++) {
		if(m_t_image_data.p_framebuffer[i])
			delete[] m_t_image_data.p_framebuffer[i];
	}
	// cleanup

	return p_bitmap;
}

bool CJpegDecoder::Read_RestartInterval(CJpegDecoder *p_this, CFileBuffer *p_file_buffer)
{
	unsigned short n_length;
	if(!p_file_buffer->Read_Short(n_length))
		return false;

#ifndef __STRIP_ERROR_CHECKS
	if(n_length < 4)
		return false;
#endif

	p_this->m_t_restart.b_enabled = true;

	if(!p_file_buffer->Read_Short(p_this->m_t_restart.n_interval))
		return false;

#ifdef __VERBOSE
	printf("read restart interval definition [n_interval=%d]\n",
		p_this->m_t_restart.n_interval);
	// debug
#endif

#ifndef __STRIP_ERROR_CHECKS
	return p_file_buffer->Skip_Bytes(n_length - 4);
#else
	return true;
#endif
}

bool CJpegDecoder::Read_SkipMarker(CJpegDecoder *p_this, CFileBuffer *p_file_buffer) // return true
{
	return true;
}

bool CJpegDecoder::Read_JFIF_Header(CJpegDecoder *p_this, CFileBuffer *p_file_buffer)
{
	unsigned short n_length;
	if(!p_file_buffer->Read_Short(n_length))
		return false;

#ifndef __STRIP_ERROR_CHECKS
	if(n_length < 16)
		return false;
#endif
	//n_length -= 2;
	// length must be >= 16

	char p_s_jfif[5];
	for(char *p_ptr = p_s_jfif, *p_end = p_s_jfif + 5; p_ptr < p_end; p_ptr ++) {
		if(!p_file_buffer->Read_Byte((unsigned char&)*p_ptr))
			return false;
	}
#ifndef __STRIP_ERROR_CHECKS
	if(p_s_jfif[4] || strcmp(p_s_jfif, "JFIF"))
		return false;
#endif
	//n_length -= 5;
	// must contain JFIF signature

	unsigned char n_version_hi, n_version_lo;
	if(!p_file_buffer->Read_Byte(n_version_hi) ||
	   !p_file_buffer->Read_Byte(n_version_lo))
		return false;
#ifndef __STRIP_ERROR_CHECKS
	if(n_version_hi != 1 || n_version_lo > 2)
		return false;
#endif
	//n_length -= 2;
	// read version, should be 1.0 to 1.2

#ifdef __VERBOSE
	printf("read header [signature=\'%s\', version=%d.%d]\n",
		p_s_jfif, n_version_hi, n_version_lo);
	// debug
#endif

	n_length -= 9; // sum of all fields (1 sub instead of 3)
	if(!p_file_buffer->Skip_Bytes(n_length))
		return false;
	// skip the rest (contains 1 byte units (0 = pixels, 1 = inches, 2 = cm),
	// 2 bytes x-density per unit, 2 bytes y-density per unit, 1 bit thumbnail width,
	// 1 bit thumbnail height, RGB 3 byte per pixel uncompressed thumbnail)

	return true;
}

bool CJpegDecoder::Read_Unsupported(CJpegDecoder *p_this, CFileBuffer *p_file_buffer) // return false
{
	return false;
}

bool CJpegDecoder::Read_Skip(CJpegDecoder *p_this, CFileBuffer *p_file_buffer) // skip the whole block
{
	unsigned short n_length;
	if(!p_file_buffer->Read_Short(n_length))
		return false;

#ifdef __VERBOSE
	printf("\tskip %d bytes\n", n_length);
	// debug
#endif

	return p_file_buffer->Skip_Bytes(n_length - 2);
}

bool CJpegDecoder::Read_SkipShort(CJpegDecoder *p_this, CFileBuffer *p_file_buffer) // skip two bytes
{
	return p_file_buffer->Skip_Bytes(2);
}

bool CJpegDecoder::Read_QuantTables(CJpegDecoder *p_this, CFileBuffer *p_file_buffer) // read quant table block (there are )
{
	unsigned short n_length;
	if(!p_file_buffer->Read_Short(n_length))
		return false;
	n_length -= 2;

	while(n_length >= 65) {
		unsigned char n_quant_table_properties;
		if(!p_file_buffer->Read_Byte(n_quant_table_properties))
			return false;
		n_length --;

#ifndef __STRIP_ERROR_CHECKS
		if((n_quant_table_properties & 0xf) < 0 || (n_quant_table_properties & 0xf) > 3)
			return false;
		// index of quantization table must be <0, 3>

		if((n_quant_table_properties >> 4) != 0 && (n_quant_table_properties >> 4) != 1)
			return false;
		// quantization table precision must be 0 (8 bit) or 1 (16 bit)
#endif

		TQuantTable *p_cur_quant_table = &p_this->m_p_quant_table[n_quant_table_properties & 0x0f];
		p_cur_quant_table->b_16_bit = (n_quant_table_properties >> 4) == 1;

#ifndef __STRIP_ERROR_CHECKS
		if(n_length < ((p_cur_quant_table->b_16_bit)? 128 : 64))
			return false;
#endif
		n_length -= (p_cur_quant_table->b_16_bit)? 128 : 64;
		// there must be enough space for table data

		for(float *p_value = p_cur_quant_table->p_value,
		   *p_end = p_cur_quant_table->p_value + 64; p_value < p_end; p_value ++) {
			if(p_cur_quant_table->b_16_bit) {
				unsigned short n_tmp;
				if(!p_file_buffer->Read_Short(n_tmp))
					return false;
				*p_value = (float)n_tmp / 256; // todo - get file with such a quant_table and see what should be here
			} else {
				unsigned char n_tmp;
				if(!p_file_buffer->Read_Byte(n_tmp))
					return false;
				*p_value = (float)n_tmp;
			}
		}
		// read table data

		for(int i = 0; i < 64; i ++) {
			int n_index = (i)? p_zig_indices[i - 1] : 0;
			//int n_index = p_zig_indices[i]; // removed 1st element
			p_cur_quant_table->p_value[i] *=
				CFastDCT8::p_PrescaleTable()[n_index / 8] *
				CFastDCT8::p_PrescaleTable()[n_index % 8];
		}
		// hack! use scale factor here and i can use faster IDCT function

#ifdef __VERBOSE
		printf("read quant table [index=%d, precision=%d]\n", n_quant_table_properties & 0x0f,
			n_quant_table_properties >> 4);
		{
			const char p_fwd_zig_indices[64] = {
				 0,  1,  5,  6, 14, 15, 27, 28,
				 2,  4,  7, 13, 16, 26, 29, 42,
				 3,  8, 12, 17, 25, 30, 41, 43,
				 9, 11, 18, 24, 31, 40, 44, 53,
				10, 19, 23, 32, 39, 45, 52, 54,
				20, 22, 33, 38, 46, 51, 55, 60,
				21, 34, 37, 47, 50, 56, 59, 61,
				35, 36, 48, 49, 57, 58, 62, 63
			};

			for(int i = 0; i < 64; i ++) {
				printf("%7.2f%c", p_cur_quant_table->p_value[p_fwd_zig_indices[i]],
					((i % 8) == 7)? '\n' : ' ');
			}
		}
		// debug
#endif
	}

	_ASSERTE(!n_length);

	return true;
}

bool CJpegDecoder::Read_HuffmanTables(CJpegDecoder *p_this, CFileBuffer *p_file_buffer) // read huffman table
{
	unsigned short n_length;
	if(!p_file_buffer->Read_Short(n_length))
		return false;
	n_length -= 2;

	while(n_length >= 17) {
		unsigned char n_huff_table_properties;
		if(!p_file_buffer->Read_Byte(n_huff_table_properties))
			return false;
		n_length --;

#ifndef __STRIP_ERROR_CHECKS
		if((n_huff_table_properties & 0xf) < 0 || (n_huff_table_properties & 0xf) > 3)
			return false;

		if((n_huff_table_properties >> 4) != 0 && (n_huff_table_properties >> 4) != 1)
			return false;
#endif

		THuffmanTable *p_cur_huff_table = &p_this->m_p_huffman_table[n_huff_table_properties &
			0xf][n_huff_table_properties >> 4];
		bool b_AC_table = (n_huff_table_properties >> 4) == 1;

		int n_tot_component_num = 0;
		{int n_index = 0, n_code_value = 0;
		for(unsigned char *p_component_num = p_cur_huff_table->n_code_num,
		   *p_end = p_cur_huff_table->n_code_num + 16; p_component_num < p_end;
		   p_component_num ++, n_index ++) {
			if(!p_file_buffer->Read_Byte(*p_component_num))
				return false;
			p_cur_huff_table->p_code[n_index] = &p_cur_huff_table->p_code_table[n_tot_component_num];
			// set pointer to code list

			p_cur_huff_table->n_min_code_value[n_index] = n_code_value;
			n_code_value += *p_component_num;
			//p_cur_huff_table->n_max_code_value[n_index] = n_code_value; // it's easy to calculate, don't store it
			n_code_value <<= 1;
			// manage minimal and maximal code values

			n_tot_component_num += *p_component_num;
		}}
		n_length -= 16;
		// read total component numbers, assign pointers

#ifndef __STRIP_ERROR_CHECKS
		if(n_length < n_tot_component_num)
			return false;
#endif
		n_length -= n_tot_component_num;
		// there must be enough space for table data

		for(unsigned char *p_component = p_cur_huff_table->p_code_table,
		   *p_end = p_cur_huff_table->p_code_table + n_tot_component_num;
		   p_component < p_end; p_component ++) {
			if(!p_file_buffer->Read_Byte(*p_component))
				return false;
		}
		// read all table components

#ifdef __VERBOSE
		printf("read huffman table [index=%d, b_AC=%d] with %d components\n",
			n_huff_table_properties & 0xf, n_huff_table_properties >> 4, n_tot_component_num);
		for(int i = 0; i < 16; i ++)
			printf("%d%c", p_cur_huff_table->n_code_num[i], (i == 15)? '\n' : ' ');
		printf("it means this table defines those symbols:\n");
		int n_code = 0;
		for(int i = 0; i < 16; i ++) {
			if(p_cur_huff_table->n_code_num[i]) {
				_ASSERTE(p_cur_huff_table->n_min_code_value[i] == n_code /*&&
					p_cur_huff_table->n_max_code_value[i] == n_code + p_cur_huff_table->n_code_num[i]*/);
				printf("\t%d bit codes %d-%d == <%d,%d): ", i + 1, n_code, n_code + p_cur_huff_table->n_code_num[i] - 1, p_cur_huff_table->n_min_code_value[i], p_cur_huff_table->n_min_code_value[i] + p_cur_huff_table->n_code_num[i]);
				for(int j = n_code; j < n_code + p_cur_huff_table->n_code_num[i]; j ++) {
					for(int n = 0; n <= i; n ++)
						printf("%d", (j >> (i - n)) & 1);
					printf(" ");
				}
				printf("\n");
			}
			n_code += p_cur_huff_table->n_code_num[i];
			n_code *= 2;
		}
		// debug
#endif
	}

	_ASSERTE(!n_length);

	return true;
}

bool CJpegDecoder::Read_FrameHeader(CJpegDecoder *p_this, CFileBuffer *p_file_buffer)
{
	unsigned short n_length;
	if(!p_file_buffer->Read_Short(n_length))
		return false;

#ifndef __STRIP_ERROR_CHECKS
	if(n_length < 8)
		return false;
#endif

	TImageHeader *p_header = &p_this->m_t_image_header;

	if(!p_file_buffer->Read_Byte(p_header->n_sample_precision) ||
	   !p_file_buffer->Read_Short(p_header->n_height) ||
	   !p_file_buffer->Read_Short(p_header->n_width) ||
	   !p_file_buffer->Read_Byte(p_header->n_component_num))
		return false;
#ifndef __STRIP_ERROR_CHECKS
	if((p_header->n_sample_precision != 8 && p_header->n_sample_precision != 16) ||
	   p_header->n_height < 0 || p_header->n_width < 0 ||
	   p_header->n_component_num == 0 || p_header->n_component_num > 4)
		return false;
#endif
	// read header
	
#ifdef __VERBOSE
	printf("read frame header [bpp=%d, width=%d, height=%d, component_num=%d]\n",
		p_header->n_sample_precision, p_header->n_width, p_header->n_height, p_header->n_component_num);
	// debug
#endif

#ifndef __STRIP_ERROR_CHECKS
	if(n_length < 8 + p_header->n_component_num * 3)
		return false;
#endif
	_ASSERTE(n_length == 8 + p_header->n_component_num * 3);
	// there should be enough space for component specification (or better exact as much space)

	unsigned char n_max_sampling_horiz = 0;
	unsigned char n_max_sampling_vert = 0;
	for(TComponentInfo *p_comp_info = p_header->p_component_info,
	   *p_end = p_header->p_component_info + p_header->n_component_num;
	   p_comp_info < p_end; p_comp_info ++) {
		unsigned char n_sampling;
		if(!p_file_buffer->Read_Byte(p_comp_info->n_component_id) ||
		   !p_file_buffer->Read_Byte(n_sampling) ||
		   !p_file_buffer->Read_Byte(p_comp_info->n_quant_table_index))
			return false;
#ifndef __STRIP_ERROR_CHECKS
		if(p_comp_info->n_component_id < 1 || p_comp_info->n_component_id > 3 ||
		   p_comp_info->n_quant_table_index > 3)
			return false;
#endif
		p_comp_info->n_sampling_horiz = n_sampling >> 4;
		p_comp_info->n_sampling_vert = n_sampling & 0xf;
		if(n_max_sampling_horiz < p_comp_info->n_sampling_horiz)
			n_max_sampling_horiz = p_comp_info->n_sampling_horiz;
		if(n_max_sampling_vert < p_comp_info->n_sampling_vert)
			n_max_sampling_vert = p_comp_info->n_sampling_vert;
#ifndef __STRIP_ERROR_CHECKS
		if((p_comp_info->n_sampling_horiz != 1 && p_comp_info->n_sampling_horiz != 2 &&
		   p_comp_info->n_sampling_horiz != 4) || (p_comp_info->n_sampling_vert != 1 &&
		   p_comp_info->n_sampling_vert != 2 && p_comp_info->n_sampling_vert != 4))
			return false;
#endif
	}
	// read component specifications

	p_header->n_macro_block_x_num = (p_header->n_width + (n_max_sampling_horiz * 8 - 1)) /
		(n_max_sampling_horiz * 8); // t_odo - here's an error
	p_header->n_macro_block_num = p_header->n_macro_block_x_num *
		((p_header->n_height + (n_max_sampling_vert * 8 - 1)) / (n_max_sampling_vert * 8));
	// number of macro-blocks (not 64-element blocks in case image has other sampling than 1x1)

#ifdef __VERBOSE
	printf("\tmacro_block_x_num=%d macro_block_y_num=%d (%d blocks)\n", p_header->n_macro_block_x_num,
		p_header->n_macro_block_num / p_header->n_macro_block_x_num, p_header->n_macro_block_num);
	printf("\tn_max_sampling_horiz=%d n_max_sampling_vert=%d\n", n_max_sampling_horiz, n_max_sampling_vert);
#endif

	for(TComponentInfo *p_comp_info = p_header->p_component_info,
	   *p_end = p_header->p_component_info + p_header->n_component_num;
	   p_comp_info < p_end; p_comp_info ++) {
		p_comp_info->n_scale_horiz = n_max_sampling_horiz / p_comp_info->n_sampling_horiz;
		p_comp_info->n_scale_vert = n_max_sampling_vert / p_comp_info->n_sampling_vert;
#ifdef __VERBOSE
		printf("\tcomponent [id=%d, sampling=%dx%d, scale=%dx%d, quant_table_index=%d]\n",
			p_comp_info->n_component_id, p_comp_info->n_sampling_horiz, p_comp_info->n_sampling_vert,
			p_comp_info->n_scale_horiz, p_comp_info->n_scale_vert, p_comp_info->n_quant_table_index);
		// debug
#endif
	}
	// determine component scaling

	p_this->m_t_image_data.b_read_image_header = true;
	for(TComponentInfo *p_comp_info = p_header->p_component_info,
	   *p_end = p_header->p_component_info + p_header->n_component_num;
	   p_comp_info < p_end; p_comp_info ++) {
		int n = p_comp_info->n_component_id - 1;
		p_this->m_t_image_data.n_framebuffer_width[n] = (p_header->n_width +
			p_comp_info->n_scale_horiz - 1) / p_comp_info->n_scale_horiz;
		p_this->m_t_image_data.n_framebuffer_height[n] = (p_header->n_height +
			p_comp_info->n_scale_vert - 1) / p_comp_info->n_scale_vert;
		if(!(p_this->m_t_image_data.p_framebuffer[n] =
		   new unsigned char[p_this->m_t_image_data.n_framebuffer_width[n] *
		   p_this->m_t_image_data.n_framebuffer_height[n]]))
			return false;
	}
	// alloc frame-buffers

	return true;
}

bool CJpegDecoder::Read_StartScan(CJpegDecoder *p_this, CFileBuffer *p_file_buffer)
{
#ifndef __STRIP_ERROR_CHECKS
	if(!p_this->m_t_image_data.b_read_image_header)
		return false;
	// must've read image header already
#endif

	unsigned short n_length;
	if(!p_file_buffer->Read_Short(n_length))
		return false;
#ifndef __STRIP_ERROR_CHECKS
	if(n_length < 6)
		return false;
#endif

	TScanHeader t_scan_head;

	if(!p_file_buffer->Read_Byte(t_scan_head.n_component_num))
		return false;
#ifndef __STRIP_ERROR_CHECKS
	if(t_scan_head.n_component_num > 3)
		return false;
#endif
	_ASSERTE(t_scan_head.n_component_num == p_this->m_t_image_header.n_component_num);
	// read number of components

#ifndef __STRIP_ERROR_CHECKS
	if(n_length < 6 + 2 * t_scan_head.n_component_num)
		return false;
#endif
	_ASSERTE(n_length == 6 + 2 * t_scan_head.n_component_num);

#ifdef __VERBOSE
	printf("start of scan [component_num=%d]\n", t_scan_head.n_component_num);
	// debug
#endif

	for(TScanHeader::TComponentInfo2 *p_comp = t_scan_head.p_component_info,
	   *p_end = t_scan_head.p_component_info + t_scan_head.n_component_num;
	   p_comp < p_end; p_comp ++) {
		unsigned char n_entropy_table;
		if(!p_file_buffer->Read_Byte(p_comp->n_component_id) ||
		   !p_file_buffer->Read_Byte(n_entropy_table))
			return false;

		p_comp->n_DC_entropy_table = n_entropy_table >> 4;
		p_comp->n_AC_entropy_table = n_entropy_table & 0xf;
#ifndef __STRIP_ERROR_CHECKS
		if(p_comp->n_DC_entropy_table > 3 || p_comp->n_AC_entropy_table > 3)
			return false;
#endif

		p_comp->p_component = 0;
		for(TComponentInfo *p_ih_comp = p_this->m_t_image_header.p_component_info,
		   *p_ih_end = p_this->m_t_image_header.p_component_info + t_scan_head.n_component_num;
		   p_ih_comp < p_ih_end; p_ih_comp ++) {
			if(p_ih_comp->n_component_id == p_comp->n_component_id) {
				p_comp->p_component = p_ih_comp;
				break;
			}
		}
#ifndef __STRIP_ERROR_CHECKS
		if(!p_comp->p_component)
			return false;
#endif
		// find the component in image header

		p_comp->p_DC_table = &p_this->m_p_huffman_table[p_comp->n_DC_entropy_table][0];
		p_comp->p_AC_table = &p_this->m_p_huffman_table[p_comp->n_AC_entropy_table][1];
		// find huffman tables

#ifdef __VERBOSE
		printf("\tcomponent [component_id=%d, DC_entropy_table=%d, AC_entropy_table=%d]\n",
			p_comp->n_component_id, p_comp->n_DC_entropy_table, p_comp->n_AC_entropy_table);
		// debug
#endif
	}
	// read components
	
	if(!p_file_buffer->Read_Byte(t_scan_head.n_spectrum_first) ||
	   !p_file_buffer->Read_Byte(t_scan_head.n_spectrum_last) ||
	   !p_file_buffer->Read_Byte(t_scan_head.n_sucessive_approximation))
		return false;
	_ASSERTE(t_scan_head.n_spectrum_first == 0 && t_scan_head.n_spectrum_last == 63);
	// read some more info (parts of it used for lossless mode, part for point transform,
	// part of it is for progressive DCT mode - progressive using saving just a part of
	// coefficients - image quality improves as more and more coefficients are read or
	// using successive approximation (saving component's integer most important bytes,
	// then lower and lower important bytes ...))

	CBitReader bit_reader; // blocks come in single bit-stream

	for(int i = 0; i < p_this->m_t_image_header.n_macro_block_num; i ++) {
		if(!p_this->Decode_MacroBlock(bit_reader, i, &t_scan_head, p_file_buffer))
			return false;
		if(p_this->m_t_restart.b_enabled && (i + 1) % p_this->m_t_restart.n_interval == 0 &&
		   i + 1 < p_this->m_t_image_header.n_macro_block_num) {
			for(int n = 0; n < 4; n ++)
				p_this->m_t_image_data.n_dc_value[n] = 0;
			// reset DC values

			unsigned short n_marker_code;
			if(!p_file_buffer->Read_Marker(n_marker_code))
				return false;
#ifndef __STRIP_ERROR_CHECKS
			if(n_marker_code < marker_RestartMod0 ||
			   n_marker_code > marker_RestartMod7)
				return false;
#endif
			// there has to be an restart marker

#ifdef __VERBOSE
			printf("\tencountered restart marker %d\n", n_marker_code - marker_RestartMod0);
			// debug
#endif

			bit_reader.Restart();
			// restart bit reader (will fetch another byte)
		}
	}
	// now the file contains image data, it needs to be decoded now ...

	return true;
}

bool CJpegDecoder::Decode_MacroBlock(CBitReader &r_bit_reader,
	int n_macro_block_index, TScanHeader *p_scan_head, CFileBuffer *p_file_buffer)
{
#ifdef __DECODE_VERBOSE
	printf("\n\t\t\t=== decode macro-block %d ===\n", n_macro_block_index + 1);
	// debug
#endif

	for(const TScanHeader::TComponentInfo2 *p_comp = p_scan_head->p_component_info,
	   *p_end = p_scan_head->p_component_info + p_scan_head->n_component_num;
	   p_comp < p_end; p_comp ++) {
		int n_x_block_num = (m_t_image_header.n_width / p_comp->p_component->n_scale_horiz + 7) / 8;
		// calc number of x blocks

		int n_raster_width = m_t_image_data.n_framebuffer_width[p_comp->n_component_id - 1]; //m_t_image_header.n_width / p_comp->p_component->n_scale_horiz;
		int n_raster_height = m_t_image_data.n_framebuffer_height[p_comp->n_component_id - 1]; //m_t_image_header.n_height / p_comp->p_component->n_scale_vert;
		// raster size

		const TQuantTable *p_quant_table = &m_p_quant_table[p_comp->p_component->n_quant_table_index];
		const THuffmanTable *p_DC_table = p_comp->p_DC_table;
		const THuffmanTable *p_AC_table = p_comp->p_AC_table;

		for(int n_block = 0, n_block_num = p_comp->p_component->n_sampling_horiz *
		   p_comp->p_component->n_sampling_vert; n_block < n_block_num; n_block ++) {
			// sampling tells us how much blocks is contained within a single scan

			float p_coeffs[64] = {0};
			// decoded coefficients; has to be de-zig-zaged

			{unsigned short n_number = 0;
			for(int i = 0; /*i < 16*/; i ++) {
				n_number <<= 1;
				n_number |= r_bit_reader.n_GetBit(p_file_buffer);
#ifndef __STRIP_ERROR_CHECKS
				if(r_bit_reader.b_Error())
					return false;
#endif
				// get another bit to our number

				if(n_number >= p_DC_table->n_min_code_value[i] &&
				   n_number < p_DC_table->n_min_code_value[i] + p_DC_table->n_code_num[i]) {
					// we have code in table

#ifdef __DECODE_VERBOSE
					printf("DC %d, ", n_number);
					// debug
#endif

					unsigned char n_huff_data = p_DC_table->p_code[i][n_number -
						p_DC_table->n_min_code_value[i]];
					// read data from huffman table, one nibble is number of preceding zeroes,
					// the other one is length of actual encoded value, in bits
					// for DC it's actualy just number of bits

					if(!n_huff_data)
						p_coeffs[0] = (float)m_t_image_data.n_dc_value[p_comp->n_component_id - 1];
					else {
						m_t_image_data.n_dc_value[p_comp->n_component_id - 1] +=
							r_bit_reader.n_GetNumber(n_huff_data, p_file_buffer);
#ifndef __STRIP_ERROR_CHECKS
						if(r_bit_reader.b_Error())
							return false;
#endif
						p_coeffs[0] = (float)m_t_image_data.n_dc_value[p_comp->n_component_id - 1];
					}

					break;
				}
#ifndef __STRIP_ERROR_CHECKS
				if(i == 15)
					return false;
#endif
			}}
			// decode DC value

			/*if(p_quant_table->b_16_bit)
				p_coeffs[0] = (float)((int)p_coeffs[0] * (p_quant_table->p_value[0] >> 8));
			else*/ // hack! - i pre-multiplied it and can use faster IDCT function, so this is no longer necessary
				p_coeffs[0] *= p_quant_table->p_value[0];
			// de-quantize DC

			{char *p_index = p_zig_indices/* + 1*/;
			for(int n = 1; n < 64;) {
				unsigned short n_number = 0;
				for(int i = 0; /*i < 16*/; i ++) {
					n_number <<= 1;
					n_number |= r_bit_reader.n_GetBit(p_file_buffer);
#ifndef __STRIP_ERROR_CHECKS
					if(r_bit_reader.b_Error())
						return false;
#endif
					// get another bit to our number

					if(n_number >= p_AC_table->n_min_code_value[i] &&
					   n_number < p_AC_table->n_min_code_value[i] + p_AC_table->n_code_num[i]) {
						// we have code in table
#ifdef __DECODE_VERBOSE
						printf("AC %d, ", n_number);
						// debug
#endif
						unsigned char n_huff_data = p_AC_table->p_code[i][n_number -
							p_AC_table->n_min_code_value[i]];
						// read data from huffman table, one nibble is number of preceding zeroes,
						// the other one is length of actual encoded value, in bits
						// for DC it's actualy just number of bits

						unsigned char n_zero_num = n_huff_data >> 4;
						unsigned char n_number_bits = n_huff_data & 0xf;

						if(!n_number_bits) {
							if(!n_zero_num) {
								/*for(char *p_end = p_zig_indices + /*64* /63; p_index < p_end;)
									p_coeffs[*p_index ++] = 0;*/
								// found eob, fill the rest of coeffs with zeroes
								p_index = p_zig_indices + /*64*/63;
#ifdef __DECODE_VERBOSE
								printf("(EOB) ");
								// debug
#endif
								n = 64; // disrupt the outer loop
								break;
							} else if(n_zero_num == 15) {
								n_zero_num ++; // the encoded zero itself
#ifndef __STRIP_ERROR_CHECKS
								if((p_index + n_zero_num) - p_zig_indices > 63)
									return false;
#endif
								/*for(char *p_end = p_index + n_zero_num; p_index < p_end;)
									p_coeffs[*p_index ++] = 0;*/
								p_index += n_zero_num;
								n += n_zero_num;

#ifdef __DECODE_VERBOSE
								printf("(16x0) ");
								// debug
#endif
							} else
								return false;
						} else {
#ifndef __STRIP_ERROR_CHECKS
							if((p_index + n_zero_num + 1) - p_zig_indices > 63) // n_zero_num + actual coeff
								return false;
#endif
							/*for(char *p_end = p_index + n_zero_num; p_index < p_end;)
								p_coeffs[*p_index ++] = 0;*/
							p_index += n_zero_num;
							n += n_zero_num;
							// fill preceding zeroes

							/*if(p_quant_table->b_16_bit) {
								p_coeffs[*p_index ++] = (float)((int)r_bit_reader.n_GetNumber(
									n_number_bits, p_file_buffer) * (p_quant_table->p_value[n] >> 8));
							} else {*/ // hack! - i pre-multiplied it and can use faster IDCT function, so this is no longer necessary
								p_coeffs[*p_index ++] = (float)r_bit_reader.n_GetNumber(n_number_bits,
									p_file_buffer) * p_quant_table->p_value[n];
							//}
							// de-quantize

#ifndef __STRIP_ERROR_CHECKS
							if(r_bit_reader.b_Error())
								return false;
#endif
							n ++;
							// read vli
						}

						break;
					}
#ifndef __STRIP_ERROR_CHECKS
					if(i == 15)
						return false;
#endif
				}
			}
			_ASSERTE(p_index == p_zig_indices + /*64*/63); // should never happen
			}
			// decode 63 AC values, dequantize as well (saving lots of muls because i'm not
			// de-quantizing zeros)
			// t_odo - is it smaller code with array initialized to zeros or with those zero-ing
			// loops? code is both smaller and faster
			// todo - is code smaller when it's zig-zagged on the fly or when it's copied into
			// another array in a loop afterwards?

#ifdef __DECODE_VERBOSE
			printf("\ndecoded block:\n");
			for(int i = 0; i < 64; i ++)
				printf("%5d%c", (int)p_coeffs[i], ((i % 8) == 7)? '\n' : ' ');
			// debug
#endif

			int n_block_x = n_block % p_comp->p_component->n_sampling_horiz;
			int n_block_y = n_block / p_comp->p_component->n_sampling_horiz;
			int n_macro_block_x = n_macro_block_index % m_t_image_header.n_macro_block_x_num;
			int n_macro_block_y = n_macro_block_index / m_t_image_header.n_macro_block_x_num;
			n_block_x += n_macro_block_x * p_comp->p_component->n_sampling_horiz;
			n_block_y += n_macro_block_y * p_comp->p_component->n_sampling_vert;
			n_block_x *= 8;
			n_block_y *= 8;
			// calc position of block in image

#ifdef __DECODE_VERBOSE
			printf("block idct [channel=%d, raster_x=%d, raster_y=%d, copy_w=%d, copy_h=%d]\n",
				p_comp->n_component_id, n_block_x, n_block_y,
				(n_raster_width - n_block_x > 8)? 8 : n_raster_width - n_block_x,
				(n_raster_height - n_block_y > 8)? 8 : n_raster_height - n_block_y);
			// debug
#endif

			int n_copy_horiz = (n_raster_width - n_block_x > 8)? 8 : n_raster_width - n_block_x;
			int n_copy_vert = (n_raster_height - n_block_y > 8)? 8 : n_raster_height - n_block_y;
			if(n_copy_horiz > 0 && n_copy_vert > 0) {
				// there are cases when part of macroblock lies beyond the image border,
				// if this component is sampled more than 1x1, there's chance this block
				// will not be necessary to compute

				float p_dest_coeffs[64];

				CFastDCT8::Inverse(p_dest_coeffs, p_coeffs);

#ifdef __DECODE_VERBOSE
			printf("\nIDCT block:\n");
			for(int i = 0; i < 64; i ++)
				printf("%6.2f%c", p_dest_coeffs[i], ((i % 8) == 7)? '\n' : ' ');
			// debug
#endif

				unsigned char *p_dest_ptr = m_t_image_data.p_framebuffer[
					p_comp->n_component_id - 1] + n_block_x + n_block_y * n_raster_width;
				// get address of dest framebuffer

				const float *p_idct_ptr = p_dest_coeffs;

				/*char p_final_coeffs[64];
				for(char *p_final_coeffs_ptr = p_final_coeffs, *p_final_coeffs_end = p_final_coeffs + 64; p_final_coeffs_ptr < p_final_coeffs_end;) {
					*p_final_coeffs_ptr ++ = (*p_idct_ptr < -128)? -128 :
						((*p_idct_ptr <= 127)? (unsigned char)*p_idct_ptr : 127);
					p_idct_ptr ++;
				}
				const char *p_idct_ptr2 = p_final_coeffs;
				for(unsigned char *p_dest_end = p_dest_ptr + n_raster_width * n_copy_vert;
				   p_dest_ptr < p_dest_end; p_idct_ptr2 += 8, p_dest_ptr += n_raster_width)
					memcpy(p_dest_ptr, p_idct_ptr2, n_copy_horiz * sizeof(char));*/
				// a little bit faster but ugly

				for(unsigned char *p_dest_end = p_dest_ptr + n_raster_width * n_copy_vert;
				   p_dest_ptr < p_dest_end; p_idct_ptr += 8 - n_copy_horiz,
				   p_dest_ptr += n_raster_width - n_copy_horiz) {
					for(unsigned char *p_dest_end2 = p_dest_ptr + n_copy_horiz; p_dest_ptr < p_dest_end2;) {
						*p_dest_ptr ++ = (*p_idct_ptr < -128)? -128 :
							((*p_idct_ptr <= 127)? (unsigned char)*p_idct_ptr : 127);
						p_idct_ptr ++;
					}
					// clamp
				}
				// write required part of idct into image
				// a tiny bottleneck here, time spent here during decoding 2304x1478 pic is ~ 0.2 sec
			}
			// IDCT
		}
	}

	return true;
}

/*
 *								=== ~CJpegDecoder ===
 */

/*
 *								=== CJpegDecoder::CFileBuffer ===
 */

const int CJpegDecoder::CFileBuffer::m_n_buffer_size = 16384;

CJpegDecoder::CFileBuffer::CFileBuffer(FILE *p_fr)
	:m_p_fr(p_fr), m_p_buffer_ptr(m_p_buffer), m_n_buffer_used(0)
{
}

bool CJpegDecoder::CFileBuffer::Read_Byte(unsigned char &r_n_byte)
{
	if(m_p_buffer_ptr >= m_p_buffer + m_n_buffer_used) {
		m_n_buffer_used = fread(m_p_buffer, sizeof(char), m_n_buffer_size, m_p_fr);
		if(!m_n_buffer_used)
			return false;
		m_p_buffer_ptr = m_p_buffer;

#ifdef __VERBOSE
		printf("have to fetch more data from file while reading byte\n");
		// debug
#endif
	}
	r_n_byte = *m_p_buffer_ptr ++;
	return true;
}

bool CJpegDecoder::CFileBuffer::Read_Short(short &r_n_short)
{
	unsigned char p_hi_lo[2];
	if(!Read_Byte(p_hi_lo[1]) || !Read_Byte(p_hi_lo[0]))
		return false;
	_ASSERTE(sizeof(short) == sizeof(p_hi_lo));
	memcpy(&r_n_short, p_hi_lo, sizeof(short));
	return true;
}

bool CJpegDecoder::CFileBuffer::Read_Marker(unsigned short &r_n_marker_code)
{
	unsigned short n_code;
	if(!Read_Short(n_code))
		return false;
	while(1) {
		bool b_byte_offset = false;
		if((n_code & 0xff00) == 0xff00) {
			if((n_code & 0xff) == 0xff)
				b_byte_offset = true;
			else if((n_code & 0xff) != 0x00) {
#ifdef __VERBOSE
				printf("found marker word 0x%04x\n", (unsigned int)n_code & 0xffff);
				// debug
#endif

				r_n_marker_code = n_code;
				return true;
			}
		} else if((n_code & 0xff) == 0xff)
			b_byte_offset = true;
		// see if we have valid marker or if at least second
		// byte of n_code could be beginning of the marker

#ifdef __VERBOSE
		printf("failed with marker word 0x%04x, trying again\n", (unsigned int)n_code & 0xffff);
		// debug
#endif

		if(b_byte_offset) {
			unsigned char n_byte;
			if(!Read_Byte(n_byte))
				return false;
			_ASSERTE((n_code & 0xff) == 0xff);
			n_code = (short)0xff00; // or n_code <<= 8, but this could be faster
			n_code |= n_byte;
			// second byte of n_code could be beginning of the marker
		} else {
			_ASSERTE((n_code & 0xff) != 0xff);
			if(!Read_Short(n_code))
				return false;
			// get next short
		}
	}
}

bool CJpegDecoder::CFileBuffer::Skip_Bytes(int n_byte_num)
{
	while(1) {
		if(m_n_buffer_used - (m_p_buffer_ptr - m_p_buffer) >= n_byte_num) {
			m_p_buffer_ptr += n_byte_num;
			return true;
		}
		// do we have enough bytes in buffer?

#ifdef __VERBOSE
		printf("have to fetch more data from file while skipping %d bytes\n", n_byte_num);
		// debug
#endif

		n_byte_num -= m_n_buffer_used - (m_p_buffer_ptr - m_p_buffer);
		m_n_buffer_used = fread(m_p_buffer, sizeof(char), m_n_buffer_size, m_p_fr);
		if(!m_n_buffer_used)
			return false;
		m_p_buffer_ptr = m_p_buffer;
		// read buffer
	}
}

/*
 *								=== ~CJpegDecoder::CFileBuffer ===
 */

/*
 *								=== CJpegDecoder::CBitReader ===
 */

CJpegDecoder::CBitReader::CBitReader()
	:m_n_bit_num(0), m_n_cur_byte(0), m_b_0xff_byte(false), m_b_error_flag(false)
{
}

unsigned char CJpegDecoder::CBitReader::n_GetBit(CFileBuffer *p_file_buffer)
{
	if(!m_n_bit_num) {
		unsigned char n_byte;
		if(!p_file_buffer->Read_Byte(n_byte)) {
			m_b_error_flag = true;
			return 0;
		}
		if(n_byte == 0x00 && m_b_0xff_byte) {
			if(!p_file_buffer->Read_Byte(m_n_cur_byte)) {
				m_b_error_flag = true;
				return 0;
			}
		} else
			m_n_cur_byte = n_byte;
		// fetch another byte, remember we need to look for 0xff00 sequence
		// (0xff is valid, 0x00 is inserted to prevent random markers to occur
		// in bit-coded stream) // fixme - it's in bit-coded stream only or it's in all blocks?

		m_b_0xff_byte = m_n_cur_byte == 0xff;
		// remember if we had 0xff again

		m_n_bit_num = 7; // save dec instr
	} else
		m_n_bit_num --;

	unsigned char n_result = m_n_cur_byte >> 7;
	// return MSB

	m_n_cur_byte <<= 1;
	// and shift a little

	return n_result;
}

short CJpegDecoder::CBitReader::n_GetNumber(int n_bit_num, CFileBuffer *p_file_buffer)
{
	_ASSERTE(n_bit_num);

	unsigned char n_first_bit;
	short n_number = (n_first_bit = n_GetBit(p_file_buffer));
	// read sign bit; either 1 = positive or 0 = negative, calc min number with n_bit_num bits 

	for(int n = n_bit_num; -- n;) {
		n_number <<= 1;
		n_number |= n_GetBit(p_file_buffer);
	}
	// read more bits

	if(!n_first_bit)
		n_number += (short)((0xffff ^ ((1 << n_bit_num) - 1)) | 1);
	// negative number

	return n_number;
}

void CJpegDecoder::CBitReader::Restart()
{
	m_n_bit_num = 0;
	m_b_0xff_byte = false;
}

/*
 *								=== ~CJpegDecoder::CBitReader ===
 */
